package de.hft.stuttgart.citydoctor2.healing;

import java.util.HashMap;
import java.util.Map;

import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import de.hft.stuttgart.citydoctor2.check.HealingID;
import de.hft.stuttgart.citydoctor2.check.HealingMethod;
import de.hft.stuttgart.citydoctor2.check.ModificationListener;
import de.hft.stuttgart.citydoctor2.check.error.PointTouchesEdgeError;
import de.hft.stuttgart.citydoctor2.check.error.RingDuplicatePointError;
import de.hft.stuttgart.citydoctor2.check.error.RingEdgeIntersectionError;
import de.hft.stuttgart.citydoctor2.datastructure.ConcretePolygon;
import de.hft.stuttgart.citydoctor2.datastructure.Edge;
import de.hft.stuttgart.citydoctor2.datastructure.Geometry;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing;
import de.hft.stuttgart.citydoctor2.datastructure.LinearRing.LinearRingType;
import de.hft.stuttgart.citydoctor2.datastructure.Polygon;
import de.hft.stuttgart.citydoctor2.datastructure.Vertex;
import de.hft.stuttgart.citydoctor2.math.Vector3d;
import de.hft.stuttgart.citydoctor2.math.graph.KDTree;

public class HealRingSelfIntError implements HealingMethod {
	
	private static final Logger logger = LogManager.getLogger(HealRingSelfIntError.class);

	@Override
	public boolean visit(RingEdgeIntersectionError err, ModificationListener l) {
		logger.info("Executing Repair for RingEdgeIntersectionError");
		// introduce a new vertex and create two rings out of one
		LinearRing r = err.getRing();
		Geometry geom = r.getParent().getParent();
		LinearRing r1 = new LinearRing(r.getType());
		LinearRing r2 = new LinearRing(r.getType());
		
		// if exterior ring -> 2 new polygons, otherwise only 2 inner rings
		if (err.getRing().getType() == LinearRingType.EXTERIOR) {
			ConcretePolygon p1 = new ConcretePolygon();
			p1.setExteriorRing(r1);
			ConcretePolygon p2 = new ConcretePolygon();
			p2.setExteriorRing(r2);
			geom.replacePolygon(err.getRing().getParent(), p1, p2);
		} else {
			Polygon parent = err.getRing().getParent();
			parent.removeInnerRing(err.getRing());
			parent.addInteriorRing(r1);
			parent.addInteriorRing(r2);
		}
		
		LinearRing currentRing = r1;
		Vertex inter = new Vertex(err.getIntersection());
		inter = checkDuplicatePoints(geom, inter);
		for (int i = 0; i < r.getVertices().size() - 1; i++) {
			Vertex v = r.getVertices().get(i);
			currentRing.addVertex(v);
			Vertex nextV = r.getVertices().get(i + 1);
			Edge e1 = err.getEdge1();
			Edge e2 = err.getEdge2();
			if (e1.formedByIgnoreDirection(v, nextV) || e2.formedByIgnoreDirection(v, nextV)) {
				currentRing.addVertex(inter);
				// switch rings
				if (currentRing == r1) {
					currentRing = r2;
				} else {
					currentRing = r1;
				}
			}
		}
		// close both rings
		r1.addVertex(r1.getVertices().get(0));
		r2.addVertex(r2.getVertices().get(0));
		
		return true;
	}
	
	private Vertex checkDuplicatePoints(Geometry geom, Vertex inter) {
		KDTree tree = new KDTree();
		for (Vertex v : geom.getVertices()) {
			tree.add(v);
		}
		Vertex newVertex = tree.getFirstNodeInRadius(inter, 0.0001);
		if (newVertex == null) {
			newVertex = inter;
		}
		return newVertex;
	}

	@Override
	public boolean visit(RingDuplicatePointError err, ModificationListener l) {
		logger.info("Executing Repair for RingDuplicatePointError");
		LinearRing r = err.getRing();
		
		// merge points together in case they are different
		Vertex duplicate1 = err.getVertex1();
		Vertex duplicate2 = err.getVertex2();
		double x = (duplicate1.getX() + duplicate2.getX()) / 2;
		double y = (duplicate1.getY() + duplicate2.getY()) / 2;
		double z = (duplicate1.getZ() + duplicate2.getZ()) / 2;
		Vector3d avgPoint = new Vector3d(x, y, z);
		duplicate1.setX(avgPoint.getX());
		duplicate1.setY(avgPoint.getY());
		duplicate1.setZ(avgPoint.getZ());
		duplicate2.setX(avgPoint.getX());
		duplicate2.setY(avgPoint.getY());
		duplicate2.setZ(avgPoint.getZ());
		
		// create 2 new rings touching in the duplicate point
		LinearRing r1 = new LinearRing(r.getType());
		LinearRing r2 = new LinearRing(r.getType());
		if (r.getType() == LinearRingType.EXTERIOR) {
			createNewPolygonsFromRings(r, r1, r2);
		} else {
			// replace inner ring with two new ones
			r.getParent().removeInnerRing(r);
			r.getParent().addInteriorRing(r1);
			r.getParent().addInteriorRing(r2);
		}
		
		
		LinearRing currentRing = r1;
		// do not use the last vertex
		for (int i = 0; i < r.getVertices().size() - 1; i++) {
			Vertex v = r.getVertices().get(i);
			if (v == duplicate1 || v == err.getVertex2()) {
				// switch rings
				if (currentRing == r2) {
					currentRing = r1;
				} else {
					currentRing = r2;
				}
			}
			currentRing.addVertex(v);
		}
		// close both rings
		r1.addVertex(r1.getVertices().get(0));
		r2.addVertex(r2.getVertices().get(0));
		
		eliminateIdenticalPointsWithDifferentInstance(err.getRing().getParent().getParent());
		return true;
	}

	private void eliminateIdenticalPointsWithDifferentInstance(Geometry geom) {
		Map<Vertex, Vertex> vertexMap = new HashMap<>();
		for (Polygon p : geom.getPolygons()) {
			checkRing(p.getExteriorRing(), vertexMap);
			for (LinearRing ring : p.getInnerRings()) {
				checkRing(ring, vertexMap);
			}
		}
	}

	private void checkRing(LinearRing ring, Map<Vertex, Vertex> vertexMap) {
		for (int i = 0; i < ring.getVertices().size(); i++) {
			Vertex v = ring.getVertices().get(i);
			if (vertexMap.containsKey(v)) {
				ring.getVertices().set(i, vertexMap.get(v));
			} else {
				vertexMap.put(v, v);
			}
		}
	}

	private void createNewPolygonsFromRings(LinearRing r, LinearRing r1, LinearRing r2) {
		Polygon p = r.getParent();
		Geometry geom = p.getParent();
		// create two new polygons
		ConcretePolygon p1 = new ConcretePolygon();
		p1.setExteriorRing(r1);
		ConcretePolygon p2 = new ConcretePolygon();
		p2.setExteriorRing(r2);
		
		// replace the old polygon with the new ones
		geom.replacePolygon(p, p1, p2);
	}
	
	@Override
	public boolean visit(PointTouchesEdgeError err, ModificationListener l) {
		logger.debug("Executing Repair for PointTouchesEdgeError");
		Edge e = err.getEdge();
		for (LinearRing r : e.getAdjacentRings()) {
			int index = -1;
			for (int i = 0; i < r.getVertices().size() - 1; i++) {
				Vertex v1 = r.getVertices().get(i);
				Vertex v2 = r.getVertices().get(i + 1);
				if (e.getOppositeVertex(v1) == v2) {
					index = i;
					break;
				}
			}
			if (index == -1) {
				StringBuilder sb = new StringBuilder("Ring does not contain edge.\n");
				sb.append("Ring: ");
				sb.append(r.getGmlId().toString());
				sb.append('\n');
				for (Vertex v : r.getVertices()) {
					sb.append(v);
					sb.append('\n');
				}
				sb.append("Edge:");
				sb.append('\n');
				sb.append(e.getFrom());
				sb.append('\n');
				sb.append(e.getTo());
				throw new IllegalStateException(sb.toString());
			}
			r.getVertices().add(index + 1, err.getVertex());
		}
		return true;
	}

	@Override
	public HealRingSelfIntError createNew() {
		return new HealRingSelfIntError();
	}

	@Override
	public HealingID getID() {
		return HealingID.R_SELF_INTERSECTION;
	}
}
